from cpython.object cimport PyObject

cdef extern from "frameobject.h":
    ctypedef struct PyFrameObject
    ctypedef struct _Py_CODEUNIT

    ctypedef class types.CodeType[object PyCodeObject]:
        cdef object co_filename
        cdef int co_firstlineno
        cdef object co_code
        cdef str co_name
        cdef int co_argcount
        cdef tuple co_varnames

    ctypedef class types.FrameType[object PyFrameObject]:
        cdef CodeType f_code
        cdef PyObject *f_trace
        cdef object f_globals
        cdef object f_locals
        cdef int f_lineno
        cdef char f_trace_lines
        cdef char f_trace_opcodes
        cdef int f_lasti
        cdef PyObject *f_valuestack
        cdef PyObject *f_stacktop

cdef extern from "frame.c":
    PyObject *get_valuestack_full(PyFrameObject *frame)
    PyObject *get_valuestack(PyFrameObject*frame, Py_ssize_t index)
    unsigned int get_arg(_Py_CODEUNIT *codestr, Py_ssize_t i)

cdef class OpcodeMetric:
    cdef readonly Py_ssize_t hits
    cdef readonly Py_ssize_t builtin_calls
    cdef Py_ssize_t get_value(self, FrameType frame)
    cdef Py_ssize_t get_function_opcodes(self, object function, object args, object kwargs)

cdef class AdvancedOpcodeMetric(OpcodeMetric):
    pass



cdef class CallMetric:
    cdef object args
    cdef object kwargs
    cdef public Py_ssize_t value(self)

cdef class ArgsLengthLinear(CallMetric):
    pass

cdef public dict AVG_BUILTIN_FUNCTION_COMPLEXITY


from cpython.object cimport PyObject_GetAttr
cdef class CallMetric:
    """
        Average case complexities

        Generally, 'n' is the number of elements currently in the container.
        'k' is either the value of a parameter or the number of elements in the parameter.

        As in:  https://wiki.python.org/moin/TimeComplexity
        TODO: How should this be weighted ? C functions are faster
            The resulting value should be somewhat equivalent to counting bytecode instructions
    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    cdef public Py_ssize_t value(self):
        raise NotImplementedError("Must be implemented by subclass.")

cdef class ArgsLengthLinear(CallMetric):
    cdef public Py_ssize_t value(self):
        if PyObject_GetAttr(self.args[0], '__iter__'):
            return len(self.args[0])
        return len(self.args)

AVG_BUILTIN_FUNCTION_COMPLEXITY = {
    max: ArgsLengthLinear,
    sorted: ArgsLengthLinear,
}



import argparse
import inspect
import logging
import os
import string
import sys
from pathlib import Path

import matplotlib.pyplot as plt
from py_pal.tracer import Tracer

from py_pal.complexity import Complexity
from py_pal.estimator import ComplexityEstimator, Columns


def check_positive(value):
    int_value = int(value)
    if int_value <= 0:
        raise argparse.ArgumentTypeError("%s is an invalid positive int value" % value)
    return int_value


def main():
    assert sys.version_info >= (3, 7)

    parser = argparse.ArgumentParser(description='Profile')
    parser.add_argument('-f', '--function', type=str, help='specify a function')
    parser.add_argument('-l', '--line', help='Calculate complexity for each line', action='store_true')
    parser.add_argument('-v', '--visualize', help='Plot runtime graphs', action='store_true')
    parser.add_argument('-s', '--separate', help='Estimate function complexity for each argument', action='store_true')
    parser.add_argument('-o', '--out', type=str, help='Output directory', default='stats')
    parser.add_argument('--save', help='Save statistics', action='store_true')
    parser.add_argument('--debug', help='Log debug output', action='store_true')
    parser.add_argument('--format', type=str, help='Output format, possible types are: csv, html, excel, json',
                        default='csv')
    parser.add_argument('target', type=str, help='a Python file or import path')

    args, unknown = parser.parse_known_args()

    logging.basicConfig(
        level=logging.DEBUG if args.debug else logging.INFO,
        format='[%(levelname)s, %(module)s, %(funcName)s]: %(message)s'
    )
    logger = logging.getLogger(__name__)

    function = None
    if args.function:
        function = getattr(__import__(args.target, fromlist=[args.function]), args.function)
        sys.argv = [inspect.getfile(function), *unknown]

    file = None
    try:
        file = open(args.target).read()
    except FileNotFoundError:
        pass

    if not function and file is None:
        raise ValueError("File or function could not be loaded")

    # TODO: Test cases: simple examples, sorting algorithms, games(pygame, term2048, mario, ...)
    # TODO: Output: function, parameter_name+nummer, complexity, line, module(filename), amount_data_points

    tracer = Tracer()

    if file:
        code = compile(file, filename=args.target, mode='exec')
        _globals = globals()

        # Execute as direct call e.g. 'python example.py'
        _globals['__name__'] = '__main__'

        # Append path to enable module import resolution in client code
        sys.path.append(os.path.dirname(args.target))

        # Pass arguments
        sys.argv = [args.target, *unknown]

        tracer.trace()
        exec(code, _globals, _globals)
        tracer.stop()

    if function:
        tracer.trace()
        function()
        tracer.stop()

    estimator = ComplexityEstimator(tracer)
    res = estimator.export(line=args.line, separate=args.separate)

    logger.debug(res[[
        Columns.FUNCTION_NAME, Columns.ARG_DEP, Columns.COMPLEXITY, Columns.LINE, Columns.FILE,
        Columns.DATA_POINTS
    ]].dropna().sort_values(by=[Columns.COMPLEXITY]).to_string())
    logger.info(res[[
        Columns.FUNCTION_NAME, Columns.ARG_DEP, Columns.COMPLEXITY, Columns.LINE, Columns.FILE,
        Columns.DATA_POINTS
    ]].to_string())

    # pd.options.display.max_colwidth = 100

    if not args.save and args.visualize:
        res = res[[isinstance(x, Complexity) for x in res[Columns.COMPLEXITY]]]

        for index, data_frame in res.iterrows():
            plot = plot_data_points(data_frame)
            plot.show()

    if args.save:
        if not os.path.isdir(args.out):
            os.mkdir(args.out)

        filename = os.path.basename(args.target)

        if args.format == 'csv':
            res.to_csv("{}\\{}.csv".format(args.out, filename))
        if args.format == 'html':
            res.to_html("{}\\{}.html".format(args.out, filename))
        if args.format == 'excel':
            res.to_excel("{}\\{}.excel".format(args.out, filename))
        if args.format == 'json':
            res.to_json("{}\\{}.json".format(args.out, filename))
        if args.visualize:
            res = res[[isinstance(x, Complexity) for x in res[Columns.COMPLEXITY]]]

            valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
            for index, data_frame in res.iterrows():
                filename = "{}_{}_{}.png".format(data_frame[Columns.FUNCTION_NAME], data_frame[Columns.ARG_DEP], index)
                filename_escaped = ''.join(c for c in filename if c in valid_chars)
                filename_escaped = filename_escaped.replace(' ', '_')
                plot = plot_data_points(data_frame)
                plot.savefig(
                    Path(args.out) / filename_escaped
                )


def plot_data_points(data_frame):
    tracing_data = data_frame[Columns.TRACING_DATA]
    current_axis = plt.gca()
    x_axis = tracing_data.columns[0]
    tracing_data.plot(
        x=x_axis,
        y=Columns.NORM_OPCODE_WEIGHT,
        kind='scatter',
        title="Function: {}, Args: {}".format(data_frame[Columns.FUNCTION_NAME], data_frame[Columns.ARG_DEP]),
        ax=current_axis
    )
    x_values = tracing_data[x_axis].to_numpy()
    current_axis.plot(
        x_values,
        data_frame[Columns.COMPLEXITY].compute(x_values),
        color='red'
    )
    return plt


if __name__ == "__main__":
    main()



import logging
from abc import ABC

import numpy as np
import pandas as pd
from numpy.linalg import LinAlgError
from pandas import DataFrame

from py_pal.complexity import ALL_CLASSES, UnderDeterminedEquation

LOGGER = logging.getLogger(__name__)


class Estimator(ABC):
    def __init__(self, tracer):
        self.tracer = tracer

    @property
    def calls(self):
        data_frame = DataFrame(
            columns=[
                Columns.CALL_ID, Columns.FUNCTION_ID, Columns.FILE, Columns.FUNCTION_LINE, Columns.FUNCTION_NAME,
                Columns.ARGS_NAMES,
                Columns.ARGS_PROXY, Columns.KWARGS_NAMES, Columns.KWARGS_PROXY, Columns.VARARGS_NAMES,
                Columns.VARARGS_PROXY, Columns.VARKW_NAMES, Columns.VARKW_PROXY,
            ],
            data=self.tracer.get_call_stats()
        )

        # Remove columns if arguments are not present
        data_frame.dropna(axis=1, how='all', inplace=True)
        LOGGER.debug(len(data_frame))
        return data_frame

    @property
    def opcodes(self):
        opcode_statistics = self.tracer.get_opcode_stats()
        if opcode_statistics.size == 0:
            return DataFrame(columns=[Columns.CALL_ID, Columns.LINE, Columns.OPCODE_WEIGHT])

        data_frame = DataFrame(
            columns=[Columns.CALL_ID, Columns.LINE, Columns.OPCODE_WEIGHT],
            data=opcode_statistics,
        )
        LOGGER.debug(len(data_frame))
        return data_frame

    # Data transformation
    def aggregate_opcodes(self, line):
        # Sum opcode weights and map them to their respective calls.
        if line:
            opcode_sum = self.opcodes.groupby(
                [Columns.CALL_ID, Columns.LINE]
            )[Columns.OPCODE_WEIGHT].agg(OpcodeWeight=np.sum).reset_index()
        else:
            opcode_sum = self.opcodes.groupby(Columns.CALL_ID)[Columns.OPCODE_WEIGHT].agg(
                OpcodeWeight=np.sum).reset_index()

        data_frame = pd.merge(self.calls, opcode_sum, on=Columns.CALL_ID)

        # Clean
        data_frame.drop_duplicates(inplace=True)
        LOGGER.debug("line=%s: %s", line, len(data_frame))
        return data_frame

    @staticmethod
    def infer_complexity(data_frame, arg_column):
        """
        Derive the big O complexity class.

        :param arg_column: Argument column to use as x-axis.
        :param data_frame: Time series-like data.
                            x-axis: argument size
                            y-axis: executed opcodes
        :return: Best fitting complexity class
        """
        best_class = None
        best_residuals = np.inf
        fitted = {}

        for class_ in ALL_CLASSES:
            inst = class_()
            residuals = inst.fit(
                data_frame[arg_column].values.astype(np.float),
                data_frame[Columns.NORM_OPCODE_WEIGHT].values.astype(np.float)
            )
            fitted[inst] = residuals

            if residuals < best_residuals - 1e-6:
                best_residuals = residuals
                best_class = inst

        return best_class


# TODO: check if complexity estimations are independent from each other (not modifying their data frames)
# TODO: handle kwargs/varargs/varkw ? what to do about all other arguments ?
class ComplexityEstimator(Estimator):
    def analyze_args_separate_ascending(self, data_frame):
        # TODO: Experimental feature make available as such
        # View separate argument axes and sort argument proxy value ascending
        for column in data_frame:
            if column == Columns.NORM_OPCODE_WEIGHT:
                continue

            # Keep influence of other arguments as low as possible by selecting rows with smallest proxy value
            args_columns = data_frame.columns.to_list()[:-1]
            data_frame = data_frame.sort_values(args_columns)
            data_frame.drop_duplicates([column, Columns.NORM_OPCODE_WEIGHT], keep='first', inplace=True)
            yield self.infer_complexity(data_frame, column), column, data_frame[[column, Columns.NORM_OPCODE_WEIGHT]]

    def analyze_all_args(self, data_frame):
        # View all argument axes together and sort argument proxy value ascending
        if Columns.ARGS_PROXY not in data_frame:
            return None, DataFrame()

        args_df = pd.DataFrame(data_frame[Columns.ARGS_PROXY].tolist(), index=data_frame.index)
        args_df['mean'] = args_df.mean(axis=1)
        args_df[Columns.NORM_OPCODE_WEIGHT] = data_frame[Columns.NORM_OPCODE_WEIGHT]

        # Take average of ArgProxy values and opcode weights
        if not args_df.empty:
            args_df = args_df.groupby('mean')[Columns.NORM_OPCODE_WEIGHT].mean().reset_index()
            args_df.sort_values('mean', inplace=True)

        return self.infer_complexity(args_df, 'mean'), args_df[['mean', Columns.NORM_OPCODE_WEIGHT]]

    @staticmethod
    def unpack_tuples(data_frame):
        # Unpack tuples of arguments
        slice = data_frame.copy()
        slice.dropna(subset=[Columns.ARGS_PROXY], inplace=True)
        unpacked_data_frame = pd.DataFrame(zip(*slice[Columns.ARGS_PROXY])).transpose()
        unpacked_data_frame[Columns.NORM_OPCODE_WEIGHT] = slice[Columns.NORM_OPCODE_WEIGHT]
        unpacked_data_frame.dropna(inplace=True, )
        return unpacked_data_frame

    @staticmethod
    def normalize_column(data_frame):
        _min = data_frame.min()
        _max = data_frame.max()
        # Replacing is needed to infer complexity
        return data_frame.apply(lambda x: (x - _min) / (_max - _min))

    def get_data_points(self, data, group_by, return_columns):
        for _, data_frame in data.groupby(group_by, sort=False):
            slice = DataFrame.copy(data_frame)
            filename = slice[Columns.FILE].values[0]
            line = slice[Columns.FUNCTION_LINE].values[0]
            function = slice[Columns.FUNCTION_NAME].values[0]
            if Columns.ARGS_NAMES in slice:
                arg_names = slice[Columns.ARGS_NAMES].values[0]
            else:
                arg_names = None

            # Normalize opcode weights for least squares
            slice[Columns.NORM_OPCODE_WEIGHT] = self.normalize_column(slice[Columns.OPCODE_WEIGHT])

            if all(map(lambda x: x in slice, return_columns)):
                data_frame_out = slice[return_columns]
            else:
                data_frame_out = DataFrame()
            yield filename, line, function, arg_names, data_frame_out

    @staticmethod
    def map_arg_names(pos, names):
        if not names:
            return ""
        if isinstance(pos, int):
            return names[pos]
        return list(map(lambda x: names[x], pos))

    def analyze_complexity(self, per_line, separate_args):
        LOGGER.info(
            '%s(per_line=%s, separate_args=%s)',
            self.analyze_complexity.__name__, per_line, separate_args
        )

        data = self.aggregate_opcodes(per_line)

        if per_line:
            iterator = self.get_data_points(
                data,
                [Columns.FUNCTION_ID, Columns.LINE],
                [Columns.ARGS_PROXY, Columns.NORM_OPCODE_WEIGHT, Columns.LINE]
            )
        else:
            iterator = self.get_data_points(
                data,
                [Columns.FUNCTION_ID],
                [Columns.ARGS_PROXY, Columns.NORM_OPCODE_WEIGHT]
            )

        for filename, line, function, arg_names, data_frame in iterator:
            LOGGER.debug("%s: %s", function, len(data_frame))
            if separate_args:
                args_df = self.unpack_tuples(data_frame)

                arg = 0
                try:
                    for best, arg_pos, data_points in self.analyze_args_separate_ascending(args_df):
                        yield filename, line, function, self.map_arg_names(arg_pos, arg_names), best, \
                              len(data_points), data_points
                except (LinAlgError, UnderDeterminedEquation) as exception:
                    yield filename, line, function, self.map_arg_names(arg, arg_names), \
                          exception.__class__.__name__, None, None
                finally:
                    arg += 1

            try:
                complexity, data_points = self.analyze_all_args(data_frame)
                yield filename, line, function, arg_names, complexity, len(
                    data_points), data_points
            except (LinAlgError, UnderDeterminedEquation) as exception:
                yield filename, line, function, arg_names, exception.__class__.__name__, None, None

    def export(self, line=False, separate=False):
        data_frame = DataFrame(
            self.analyze_complexity(line, separate),
            columns=[Columns.FILE, Columns.LINE, Columns.FUNCTION_NAME, Columns.ARG_DEP, Columns.COMPLEXITY,
                     Columns.DATA_POINTS, Columns.TRACING_DATA, ]
        )

        # Reorder columns
        cols = [Columns.FUNCTION_NAME, Columns.ARG_DEP, Columns.COMPLEXITY, Columns.LINE, Columns.FILE,
                Columns.DATA_POINTS, Columns.TRACING_DATA]
        data_frame = data_frame[cols]

        return data_frame


class PerformanceTestingAnalyzer(Estimator):
    def report(self):
        raise NotImplementedError()


class Columns:
    """Column names for the DataFrame objects used in the estimator."""
    CALL_ID = 'CallID'
    FILE = 'File'
    FUNCTION_NAME = 'FunctionName'
    ARGS_PROXY = 'ArgsProxy'
    ARGS_NAMES = 'Args'
    KWARGS_PROXY = 'KwargsProxy'
    KWARGS_NAMES = 'Kwargs'
    VARARGS_PROXY = 'VarArgsProxy'
    VARARGS_NAMES = 'VarArgs'
    VARKW_PROXY = 'VarKwProxy'
    VARKW_NAMES = 'VarKw'
    LINE = 'Line'
    FUNCTION_LINE = 'FunctionLine'
    OPCODE_WEIGHT = 'OpcodeWeight'
    NORM_OPCODE_WEIGHT = 'NormOpcodeWeight'
    ARG_DEP = 'DependentArgs'
    COMPLEXITY = 'Complexity'
    DATA_POINTS = 'DataPoints'
    TRACING_DATA = 'TracingData'
    FUNCTION_ID = 'FunctionID'



import cython
from cpython.pystate cimport Py_tracefunc

from py_pal.opcode_metric cimport OpcodeMetric

cdef extern from "frameobject.h":
    ctypedef struct PyObject

    ctypedef class types.CodeType[object PyCodeObject]:
        cdef object co_filename
        cdef int co_firstlineno
        cdef object co_code
        cdef str co_name
        cdef int co_argcount
        cdef tuple co_varnames

    ctypedef class types.FrameType[object PyFrameObject]:
        cdef CodeType f_code
        cdef PyObject *f_trace
        cdef object f_globals
        cdef object f_locals
        cdef int f_lineno
        cdef char f_trace_lines
        cdef char f_trace_opcodes
        cdef int f_lasti
        cdef PyObject *f_valuestack
        cdef PyObject *f_stacktop
        #cdef PyObject *f_localsplus[1]

    void PyEval_SetTrace(Py_tracefunc func, PyObject *obj)
    void PyEval_SetProfile(Py_tracefunc func, PyObject *obj)


@cython.final
cdef class Tracer:
    cdef:
        Py_ssize_t call_id
        dict opcodes
        list calls
        dict f_weight_map
        readonly dict f_map
        list call_stack
        list blacklist
        list functions
        OpcodeMetric metric

    cpdef object get_call_stats(self)
    cpdef object get_opcode_stats(self)


cdef extern from "frameobject.h":
    void PyFrame_FastToLocals(FrameType frame)



from functools import wraps
from sys import gettrace

from py_pal.tracer import Tracer

OPCODES = []
CALLS = []


def profile(function_to_trace=None, **trace_options):
    def tracing_decorator(func):
        @wraps(func)
        def tracing_wrapper(*args, **kwargs):
            if gettrace():
                return func(*args, **kwargs)

            tracer = Tracer(**trace_options)
            tracer.trace()
            try:
                return func(*args, **kwargs)
            finally:
                tracer.stop()
                OPCODES.append((tracer.get_opcode_stats()))
                CALLS.append((tracer.get_call_stats()))

        return tracing_wrapper

    if function_to_trace is None:
        return tracing_decorator
    return tracing_decorator(function_to_trace)



import io
import os
import pathlib
import re
from os.path import dirname
from os.path import join

from setuptools import setup, find_packages, Extension


def read(*names, **kwargs):
    with io.open(
            join(dirname(__file__), *names),
            encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


try:
    # Allow installing package without any Cython available. This
    # assumes you are going to include the .c files in your sdist.
    import Cython
except ImportError:
    Cython = None

compiler_options = {
    'language_level': '3str',
    'embedsignature': True,
}
macros = []
env_trace = os.environ.get('PYPAL_TRACE')
if env_trace:
    # Add directives for Cython code coverage
    compiler_options.update({'linetrace': True})
    macros = [('CYTHON_TRACE', '1')]
else:
    compiler_options.update({
        'boundscheck': False,
        'wraparound': False,
        'infer_types': True,
    })

ext = '.pyx' if Cython else '.c'
extensions = [
    Extension(
        'py_pal.tracer',
        sources=['src/py_pal/tracer' + ext],
        define_macros=macros,
    ),
    Extension(
        'py_pal.opcode_metric',
        sources=['src/py_pal/opcode_metric' + ext],
        define_macros=macros,
        include_dirs=['src/py_pal']
    ),
    Extension(
        'py_pal.metric',
        sources=['src/py_pal/metric' + ext],
        define_macros=macros,
        include_dirs=['src/py_pal']
    ),
]

if Cython:
    from Cython.Build import cythonize

    extensions = cythonize(extensions, compiler_directives=compiler_options)

# The directory containing this file
HERE = pathlib.Path(__file__).parent

if os.environ.get('CI_COMMIT_TAG'):
    version = os.environ['CI_COMMIT_TAG']
else:
    version = "0.0.1"

# This call to setup() does all the work
setup(
    name='py-pal',
    version=version,
    description='Estimate Asymptotic Runtime Complexity from Bytecode executions',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
        read('CHANGELOG.rst')
    ),
    long_description_content_type='text/x-rst',
    url='https://github.com/segroup-uni-trier/jung-ma-asymptotic-performance-testing.git',
    author='Lukas Jung',
    author_email='mail@lukasjung.de',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Utilities',
        'Topic :: Software Development :: Debuggers',
    ],
    python_requires='>=3.7',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'pandas>=0.25',
        'numpy>=1.16',
        'matplotlib>=3.1'
    ],
    entry_points={
        'console_scripts': [
            'pypal=py_pal.__main__:main',
        ]
    },
    setup_requires=[
        'cython>=0.29.14',
    ] if Cython else [],
    ext_modules=extensions
)



from collections import deque

import numpy as np
from py_pal.opcode_metric import AdvancedOpcodeMetric, OpcodeMetric
from py_pal.opcode_metric cimport OpcodeMetric

cdef tuple WHAT_NAMES = ("call", "exception", "line", "return", "c_call", "c_exception", "c_return", "opcode")

cdef class Tracer:
    def __init__(self, OpcodeMetric metric=AdvancedOpcodeMetric()):
        self.blacklist = [
            "py_pal.tracer",
            "importlib._bootstrap",
            "importlib._bootstrap_external"
        ]
        self.metric = metric
        self.call_id = 0
        self.calls = [(self.call_id, 0, '__main__', 0, '<module>', None, None, None, None, None, None, None, None)]
        self.opcodes = {}
        self.f_weight_map = {}
        self.call_stack = [(self.call_id, 0)]
        self.call_id += 1

    def __call__(self, frame, what, arg):
        if frame.f_globals.get('__name__', '') not in self.blacklist:
            # Do not measure inside of tracing machinery
            PyEval_SetTrace(<Py_tracefunc> trace_func, <PyObject *> self)

        return self

    def trace(self):
        PyEval_SetTrace(<Py_tracefunc> trace_func, <PyObject *> self)
        return self

    def stop(self):
        PyEval_SetTrace(NULL, NULL)

    cpdef get_call_stats(self):
        return np.asarray(self.calls)

    cpdef get_opcode_stats(self):
        return np.asarray([(*k, v) for k, v in self.opcodes.items()])

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

cdef int last_lineno

cdef int trace_func(Tracer self, FrameType frame, int what, PyObject *arg) except -1:
    """
    The events are emitted after opcode execution.
    Use the call and return events to structure the calls into a hierarchy.
    """
    frame.f_trace_opcodes = 1
    frame.f_trace_lines = 0

    global last_lineno

    #if frame.f_globals.get('__name__', '') in self.blacklist:
    #    return 0

    # if what == 1:
    #    print(<str> arg)
    # self.stop()

    if what == 0:
        # event: call

        # Add call as row (module, function, args, kwargs, children)
        self.call_stack.append((self.call_id, last_lineno))

        PyFrame_FastToLocals(frame)
        _args, kwonlyargs, _varargs, _varkw = _getfullargs(frame.f_code)

        if isinstance(_args, list):
            _args = tuple(_args)

        if isinstance(kwonlyargs, list):
            kwonlyargs = tuple(kwonlyargs)

        if isinstance(_varargs, list):
            _varargs = tuple(_varargs)

        if isinstance(_varkw, list):
            _varkw = tuple(_varkw)

        args = tuple(map(lambda x: frame.f_locals[x], _args)) if _args else ()
        kwargs = tuple(map(lambda x: frame.f_locals[x], kwonlyargs)) if kwonlyargs else ()
        varargs = frame.f_locals[_varargs] if _varargs else ()
        varkw = frame.f_locals[_varkw].values() if _varkw else ()

        self.calls.append((
            self.call_id,
            id(frame.f_code),
            frame.f_code.co_filename,
            frame.f_lineno,
            frame.f_code.co_name,
            _args,
            tuple(map(lambda x: get_input_factor(x), args)) if args else None,
            kwonlyargs,
            tuple(map(lambda x: get_input_factor(x), kwargs)) if kwargs else None,
            _varargs,
            tuple(map(lambda x: get_input_factor(x), varargs)) if varargs else None,
            _varkw,
            tuple(map(lambda x: get_input_factor(x), varkw)) if varkw else None
        ))
        self.call_id += 1

    elif what == 3:
        # event: return
        if len(self.call_stack) > 1:
            child = self.call_stack.pop()

        # Add opcode weight to parent call
        parent = self.call_stack[len(self.call_stack) - 1]
        value = self.f_weight_map.get(child[0], 0)
        parent_weight = self.opcodes.get(parent, 0)
        self.opcodes.update({parent: parent_weight + value})

        _value = self.f_weight_map.get(parent[0], 0)
        self.f_weight_map.update({parent[0]: _value + value})

    elif what == 7:
        # event: opcode
        # Anything in here should cause minimal overhead
        last_lineno = frame.f_lineno

        # Save opcode weight per line in current call
        call = self.call_stack[len(self.call_stack) - 1][0]
        key = (call, frame.f_lineno)
        value_line = self.opcodes.get(key, 0)
        self.opcodes.update({key: value_line + 1})

        # Keep track of all opcodes executed within call
        value = self.f_weight_map.get(call, 0)
        self.f_weight_map.update({call: value + self.metric.get_value(frame)})

        return 0

    return 0

cdef Py_ssize_t get_input_factor(arg):
    """
    Proxy for input arguments. Is used to infer complexity with the least squares algorithm.
    Therefore all returned values have to be positive and greater than zero.
    """
    if arg is None or NULL or isinstance(arg, (bool, np.bool)):
        return 1

    elif isinstance(arg, np.ndarray):
        return sum(arg.shape)

    elif isinstance(arg, np.generic):
        value = abs(int(arg))
        return value if value > 0 else 1

    elif isinstance(arg, int):
        # The memory size of an ``int`` object is constant.
        # Return ``int`` value to be able to derive simple algorithms.
        if arg == 0:
            return 1
        return abs(arg)

    elif isinstance(arg, (list, dict, set, deque, str, tuple)):
        # Length of collections
        l = len(arg)
        return l if l > 0 else 1

    else:
        return sizeof(arg)

cdef _getfullargs(co):
    nargs = co.co_argcount
    names = co.co_varnames
    nkwargs = co.co_kwonlyargcount
    args = list(names[:nargs])
    kwonlyargs = list(names[nargs:nargs + nkwargs])

    nargs += nkwargs
    varargs = None
    if co.co_flags & 4:
        varargs = co.co_varnames[nargs]
        nargs = nargs + 1
    varkw = None
    if co.co_flags & 8:
        varkw = co.co_varnames[nargs]
    return args, kwonlyargs, varargs, varkw



import types
from opcode import HAVE_ARGUMENT, EXTENDED_ARG

from py_pal.metric cimport AVG_BUILTIN_FUNCTION_COMPLEXITY, CallMetric

cdef class OpcodeMetric:
    def __init__(self):
        self.hits = 0
        self.builtin_calls = 0

    cdef Py_ssize_t get_value(self, FrameType frame):
        return 1

    cdef Py_ssize_t get_function_opcodes(self, object function, object args, object kwargs):
        cdef CallMetric instance

        if isinstance(function, types.BuiltinFunctionType):
            # Statistics
            self.builtin_calls += 1
            if function in AVG_BUILTIN_FUNCTION_COMPLEXITY:
                self.hits += 1

        if function in AVG_BUILTIN_FUNCTION_COMPLEXITY:
            complexity_class = AVG_BUILTIN_FUNCTION_COMPLEXITY[function]
            instance = complexity_class(*args, **kwargs)
            return instance.value()

        return 1

cdef unpack_oparg(code):
    # Code from dis.py
    cdef int extended_arg = 0
    for i in range(0, len(code), 2):
        op = code[i]
        if op >= HAVE_ARGUMENT:
            arg = code[i + 1] | extended_arg
            extended_arg = (arg << 8) if op == EXTENDED_ARG else 0
        else:
            arg = None
        return arg

cdef class AdvancedOpcodeMetric(OpcodeMetric):
    cdef Py_ssize_t get_value(self, FrameType frame):
        if frame.f_lasti < 0:
            return 1

        code = frame.f_code.co_code
        op = code[frame.f_lasti]

        if op == 131:
            # CALL_FUNCTION
            argc = unpack_oparg(code[frame.f_lasti:])
            valuestack = <list> get_valuestack(<PyFrameObject*> frame, argc + 1)
            args = valuestack[1:]
            _callable = valuestack[0]

            return self.get_function_opcodes(_callable, args, {})

        elif op == 141:
            # CALL_FUNCTION_KW
            argc = unpack_oparg(code[frame.f_lasti:])
            valuestack = <list> get_valuestack(<PyFrameObject*> frame, argc + 2)

            _callable = valuestack.pop(0)
            kw_names = valuestack.pop()
            kwargs = {}
            for name in kw_names:
                kwargs[name] = valuestack.pop()

            return self.get_function_opcodes(_callable, valuestack, kwargs)

        return 1



"""Definition of complexity classes."""

import numpy as np
from numpy import ma


class NotFittedError(Exception):
    pass


class UnderDeterminedEquation(Exception):
    pass


class Complexity:
    """ Abstract class that fits complexity classes to timing data."""

    def __init__(self):
        # list of parameters of the fitted function class as returned by the
        # last square method np.linalg.lstsq
        self.coeff = None

    def fit(self, n, t):
        """ Fit complexity class parameters to timing data.

        Input:
        ------

        n -- Array of values of N for which execution time has been measured.

        t -- Array of execution times for each N in seconds.

        Output:
        -------

        residuals -- Sum of square errors of fit
        """
        x = self._transform_n(n)
        y = self._transform_y(t)

        coeff, residuals, rank, s = np.linalg.lstsq(x, y, rcond=-1)

        if len(residuals) == 0:
            raise UnderDeterminedEquation

        self.coeff = coeff
        return residuals[0]

    def compute(self, n):
        """ Compute the value of the fitted function at `n`. """
        if self.coeff is None:
            raise NotFittedError()

        # Result is linear combination of the terms with the fitted coefficients
        x = self._transform_n(n)
        tot = 0
        for i in range(len(self.coeff)):
            tot += self.coeff[i] * x[:, i]
        return tot

    def __str__(self):
        prefix = '{}: '.format(self.__class__.__name__)

        if self.coeff is None:
            return prefix + ': not yet fitted'
        return prefix + self.format_str().format(*tuple(self.coeff))

    # --- abstract methods

    @classmethod
    def format_str(cls):
        """ Return a string describing the fitted function.

        The string must contain one formatting argument for each coefficient.
        """
        return 'FORMAT STRING NOT DEFINED'

    def _transform_n(self, n):
        """ Terms of the linear combination defining the complexity class.

        Output format: number of Ns x number of coefficients .
        """
        raise NotImplementedError()

    def _transform_y(self, t):
        """ Transform time as needed for fitting.
        (e.g., t->log(t)) for exponential class.
        """
        return t

    def __gt__(self, other):
        if self.__class__ == other.__class__ and self.coeff is not None and other.coeff is not None:
            return self.coeff[-1] > other.coeff[-1]
        return ALL_CLASSES.index(self.__class__) > ALL_CLASSES.index(other.__class__)


# --- Concrete implementations of the most popular complexity classes

class Constant(Complexity):
    def _transform_n(self, n):
        return np.ones((len(n), 1))

    @classmethod
    def format_str(cls):
        return '= {:.2G}'


class Linear(Complexity):
    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), n]).T

    @classmethod
    def format_str(cls):
        return '= {:.2G} + {:.2G}*n'


class Quadratic(Complexity):
    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), n * n]).T

    @classmethod
    def format_str(cls):
        return '= {:.2G} + {:.2G}*n^2'


class Cubic(Complexity):
    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), n ** 3]).T

    @classmethod
    def format_str(cls):
        return '= {:.2G} + {:.2G}*n^3'


class Logarithmic(Complexity):
    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), np.log(n)]).T

    @classmethod
    def format_str(cls):
        return '= {:.2G} + {:.2G}*log(n)'


class Linearithmic(Complexity):
    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), n * np.log(n)]).T

    @classmethod
    def format_str(cls):
        return '= {:.2G} + {:.2G}*n*log(n)'


class Polynomial(Complexity):
    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), np.log(n)]).T

    def _transform_y(self, t):
        masked = ma.log(t)
        return masked.filled(0)

    @classmethod
    def format_str(cls):
        return '= {:.2G} * x^{:.2G}'


class Exponential(Complexity):
    def _transform_n(self, n):
        return np.vstack([np.ones(len(n)), n]).T

    def _transform_y(self, t):
        masked = ma.log(t)
        return masked.filled(0)

    @classmethod
    def format_str(cls):
        return '= {:.2G} * {:.2G}^n'


ALL_CLASSES = [Constant, Logarithmic, Linear, Linearithmic, Quadratic, Cubic, Polynomial, Exponential]
